#!/usr/bin/perl
# 
# ABCpadFetcher: parse the data in web pages from
# http://www.abc.net.au/tv/guide/data/pad_output.xml and construct suitable
# data structures which can be converted into slice files for the TiVo.
#
# Copyright (c) Warren Toomey, wkt@tuhs.org, 2004, GPL license
#
#    $Revision: 1.9 $
#
# This file is distributed under the GNU Public License with the
# following additional limitations.
# 
# 1. This software must not be used in those parts of the world where
#    TiVo Inc. provides a subscription service.
# 2. This software must not be used for commercial purposes. You may 
#    not use this software to provide a "service" and charge for that
#    service.
#
# See Docs/COPYING for the GNU Public License details.
#
package TiVo::ABCpadFetcher;
use strict;
use TiVo::ConfigFiles;
use TiVo::MiscUtil;
use HTTP::Lite;

BEGIN {
  use Exporter	 ();
  use vars	 qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

  $VERSION = do { my @r = (q$Revision: 1.9 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; # must be all one line, for MakeMaker

  @ISA	       = qw(Exporter);
  @EXPORT      = qw(Fetch_ABC_Channel);
  %EXPORT_TAGS = ( );
  @EXPORT_OK   = qw();
}

use vars	 @EXPORT_OK;
END { }	 # module clean-up code here (global destructor)


#####################################
###### Package-private Globals ######
#####################################

# URL used for ABC.
my ($ABCUrl)= "http://www.abc.net.au/tv/guide/data/pad_output.xml";

# The hash of program data that is
# built up.
my (%ABCData);

# A lookup hash to determine the
# TiVo daynumber given the ABC daynumber
my (%ABCtoTiVoDay);

# Conversion from ABC genres to TiVo ones
my (%GenreTranslate) = (
	"Arts/Culture" => "ArtsMusicLiving",
	"Childrens" => "Children,ChildrensGroup",
	"Comedy" => "Comedy,ComedyGroup",
	"Current Affairs" => "PublicAffairs,NewsBusinessWeather",
	"Documentary" => "Documentary,DocumentaryGroup",
	"Drama" => "Drama,DramaGroup",
	"Education" => "Educational,EducationalGroup",
	"Factual" => "Documentary,DocumentaryGroup",
	"Movies" => "Movie,Movies",
	"Nat. Hist./Env." => "Animals,DocumentaryGroup",
	"News" => "News,NewsBusinessWeather",
	"Religion/Ethics" => "Religion",
	"Science/Technology" => "Science,ScienceNatureTechnology",
	"Sport" => "Sports,SportsGroup",
);

# Table to convert the POSIX
# timezones into ABC state names
my (%ABCState) = (
	"Australia/Sydney" => "ACT",
	"Australia/Darwin" => "NT",
	"Australia/Brisbane" => "QLD",
	"Australia/Adelaide" => "SA",
	"Australia/Hobart" => "TAS",
	"Australia/Melbourne" => "VIC",
	"Australia/Hobart" => "WA"
);

######################
###### Routines ######
######################

# Given a daynumber relative to today (0, 1, 2 etc.), return the
# dayname as a string like yyymmdd, as used in the ABC xml file.
sub Get_ABC_Dayname($)
{
  my ($wanteday)= $_[0];
  my ($dayname);
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst);

  # We want a dd/mm/yy format.
  ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)= Get_Daylist($wanteday);
  $mon++; $year+= 1900;
  $dayname= sprintf("%04d%02d%02d", $year, $mon, $mday);
  return($dayname);
}


# Given a channel name, the relative start and
# end days, fetch the web data for this channel.
# and save it. Then parse the data into the
# intermediate format. Returns 1 on no errors,
# 0 if errors.
sub Fetch_ABC_Channel($$$)
{
  my ($chan,$snum,$enum)= @_;
  my ($result1,$result2)= (1,1);
  $result1=Fetch_and_Store_Data($chan,$snum,$enum);
  $result2=Parse_Data($chan,$snum,$enum);
  return($result1 && $result2);
}

sub Fetch_and_Store_Data($$$)
{
  my ($chan,$snum,$enum)= @_;
  my ($web_proxy, $webdelay, $wanteday);
  my ($html_file, $url, $dayname);
  my ($try_again, $try, $http, $req, $body);

  # Get the output filename
  $html_file= "$ConfigVar{Webdir}/ABC_data.xml";

  # If the output file is older than 6 days, fetch a new one.
  # The pad_output.xml seems to contain 4 weeks of data, so
  # fetching it weekly means it always has enough data in it.
  if ( (-f $html_file) && (-M $html_file)<=6) {
    Debug("Using cached data ABC_data.xml\n");
  } else {

    # See if we need a web proxy, and # of times to try again
    $web_proxy= $ConfigVar{Webproxy};
    $try_again= $ConfigVar{RepeatWebConnect};

    Debug("Fetching ABC data for four weeks\n");
    $url= $ABCUrl;

    $http = new HTTP::Lite;
    if (defined($web_proxy)) { $http->proxy($web_proxy); }

    # Try 1 or more times to get the file
    if (!defined($try_again)) { $try_again=1; }

    for ($try=$try_again;$try>0;$try--) {
      my($req) = $http->request($url) && last;
    }

    if ($try==0) {
      print(STDERR "Unable to get data from ABC: $!\n");
      return(0);
    }

    # Now write the HTML document straight out
    unless (open(OUT, ">$html_file")) {
      print(STDERR "Unable to save ABC data to $html_file: $!\n");
      return(0);
    }
    print(OUT $http->body());
    close(OUT);
  }

  # Before we leave, create dummy Webdir files. This
  # prevents other fetchers from re-fetching data that
  # we already have.
  for ($wanteday= $snum; $wanteday <= $enum; $wanteday++) {
    $dayname= Get_Dayname($wanteday);
    $html_file= "$ConfigVar{Webdir}/$chan" . "_$dayname";
    Rehtml("Dummy ABC fetcher file", $html_file, 0);
  }
  return(1);
}


# Given the channel, start day and end day, read the webfiles and
# convert into intermediate format. We now put all of one day, and
# only that day, into an output file.
# New overall strategy: read in each file and parse it.
# Build up a list of records with no duration. Once the
# files are read in, go back and deduce the duration
# by looking at the differences between start times.
# Finally, write the files out.
#
sub Parse_Data($$$)
{
  my ($chan,$snum,$enum)= @_;
  my (%IgnoreDay, $urlnoamp);
  my ($wanteday, $outfile, $result);
  my ($dayname, $daynum, $pdindex, $time);
  my ($abcdaynum, $endtime, $pad);

  $result=1; undef(%ABCData);

  # Build a lookup table to find the TiVo's day from the ABC day numbers
  for ($wanteday= $snum; $wanteday <= $enum; $wanteday++) {
    $abcdaynum= Get_ABC_Dayname($wanteday);
    $daynum=  Get_Daynumber($wanteday);
    $ABCtoTiVoDay{$abcdaynum}=$daynum;
  }

  # Build the list of records for every day
  $result &= Parse_Data_For_Day($chan);

  # Now write the results out for each day where there is no data yet
  for ($wanteday= $snum; $wanteday <= $enum; $wanteday++) {
    if (defined($IgnoreDay{$wanteday})) { next; }

    # Print out the relevant records for this daynum
    $dayname= Get_Dayname($wanteday);
    $daynum=  Get_Daynumber($wanteday);

    $pdindex=0; undef(@ProgramData);
    $outfile=	"$ConfigVar{Datadir}/$chan" . "_$dayname";

    # Skip if we have already generated the output file
    if (-f $outfile) {
      Debug("Intermediate file $outfile exists, not overwriting\n");
      next;
    }

    # Skip if the hash doesn't exist
    if (!defined($ABCData{$daynum})) {
      Debug("No ABC data for $daynum $chan $dayname\n");
      print(STDERR "No ABC data for $chan on $dayname\n");
      $result=0;
    }

    # Build up a time-increasing set of records
    foreach $time (sort { $a <=> $b }
				keys( %{ $ABCData{$daynum} })) {
      $ProgramData[$pdindex]= $ABCData{$daynum}{$time};

      # See if there is a gap between this program and the previous one.
      # If there is, pad out the previous program.
      if ($pdindex>0) {
	$endtime= $ProgramData[$pdindex-1]{time} +
					$ProgramData[$pdindex-1]{duration};
        if ($endtime<$time) {
	   $pad= $time - $endtime;
	   Debug("Padding time of $ProgramData[$pdindex-1]{title} by $pad\n");
  	   $ProgramData[$pdindex-1]{duration}+= $pad;
	}
      }
      $pdindex++;
    }
    Debug("Writing file $outfile\n");
    $result &= Write_IntermediateFile($outfile);
  }
  return($result);
}

# Convert a 4-digit 24-hour time string
# into seconds from midnight
sub TwenFourToSecs($)
{
  my ($intime)=$1;
  my ($hr,$mn,$secs);
  $hr= substr($intime,0,2);
  $mn= substr($intime,2,2);
  $secs= 3600 * $hr + 60 * $mn;
  return($secs);
}

# Parse a web file for only a single day.
sub Parse_Data_For_Day($)
{
  my ($chan)= @_;
  my ($html_file);
  my ($abcdaynum, $daynum, $time, $abcstate, $tvrating, $actor, $director);
  my ($title, $duration, $description, $year, $genre, $episode, $state);
  my ($actorline, $lasttime, $endtime, $premiere, $bits);

  # Get the ABC state prefix for this channel
  $abcstate= $ABCState{ $SourceTimezone{$chan} };
  Debug("ABC state for $chan is $abcstate\n") if (defined($abcstate));

  # Get the input filename
  $html_file= "$ConfigVar{Webdir}/ABC_data.xml";

  unless (open(IN, $html_file)) {
    print(STDERR "Can't open xml file $html_file: $!\n");
    return(0);
  }
  Debug("Parsing ABC data from $html_file\n");

  while (<IN>) {
    chop;

    # Search for the beginning of a program by finding <program_detail>
    #
    next if (! /<program_detail>/);

    # Loop collecting data for this program
    undef($title); undef($duration); undef($description); undef($tvrating);
    undef($year); undef($genre); undef($episode); undef($state);
    undef($actor); undef($director); undef($premiere); $bits=0;
    while (<IN>) {
      chop;
      # Move to next record if we hit the end of this one
      if (/<\/program_detail>/) {

	# Ignore programs if they are not in our state
	if (defined($abcstate) && defined($state) &&
				($state ne "Netw") && ($state ne $abcstate)) {
	  Debug("Skipping $title as $state not equal $abcstate\n");
	  last;
	}

	# Deal with identical start times here.
 	# Always overwrite the record, as we have
	# already excluded out-of-state records above
	if (defined($ABCData{$daynum}{$time})) {
	  Debug("Overwriting duplicate date for $title in $state\n");
        }

	# The ABC data has an awful bug. Programs like
	# Compass are split into the local content (usually the
	# introduction) and the overseas content. We try to
	# merge them, but this is a hack at best.
	# Merge if the previous episode title starts with this title
	if (defined($ABCData{$daynum}{$lasttime})) {
	  if ($ABCData{$daynum}{$lasttime}{episode}=~ /^\Q$title\E/) {
	    Debug("Looks like we need to merge " .
		$ABCData{$daynum}{$lasttime}{title} . " and $title\n");
	    # Determine the end time of this program, and thus
	    # fix the duration of the last program
	    $endtime= $time+$duration;
	    $ABCData{$daynum}{$lasttime}{duration}= $endtime -
					$ABCData{$daynum}{$lasttime}{time};
	    last;
	  }
	}

	Debug("Saving $title at time $time\n");
	$ABCData{$daynum}{$time}{daynum}= $daynum;
	$ABCData{$daynum}{$time}{time}= $time;
	$ABCData{$daynum}{$time}{title}= $title;
	$ABCData{$daynum}{$time}{duration}= $duration;
	$ABCData{$daynum}{$time}{description}= $description;
	$ABCData{$daynum}{$time}{year}= $year;
	$ABCData{$daynum}{$time}{genre}= $genre;
	$ABCData{$daynum}{$time}{episode}= $episode;
	$ABCData{$daynum}{$time}{tvrating}= $tvrating;
	$ABCData{$daynum}{$time}{actor}= $actor;
	$ABCData{$daynum}{$time}{director}= $director;
	$ABCData{$daynum}{$time}{bits}= $bits if ($bits>0);
	$ABCData{$daynum}{$time}{premiere}= $premiere;
        $lasttime= $time; last;
      }

      if (/<pals_date>(\d+)<\/pals_date>/) {
	$abcdaynum=$1;
	# Ignore this record if its not in our day list
	last if (!defined($ABCtoTiVoDay{$abcdaynum}));
	$daynum= $ABCtoTiVoDay{$abcdaynum};
	next;
      }
      if (/<pals_category><..CDATA.(.*)]]><\/pals_category>/) {
	$genre= $GenreTranslate{$1} if (defined($GenreTranslate{$1}));
	next;
      }
      if (/<pad_long_title_1><..CDATA.(.*)]]><\/pad_long_title_1>/) {
	$title= $1;
	# Trim off any parentheses at the end of the title
	$title=~ s/\(.*\)$//;
	$title=~ s/\s+$//;
	Debug("Found title $1\n");
	next;
      }
      if (/<pad_long_title_2><..CDATA.(.+)]]><\/pad_long_title_2>/) {
	$episode= $1;
	next;
      }
      if (/<pals_censorship><..CDATA.(.+)]]><\/pals_censorship>/) {
	$tvrating= 3 if ($1=~ /^G/);
	$tvrating= 4 if ($1=~ /^PG/);
	$tvrating= 6 if ($1=~ /^M/);
	next;
      }
      if (/<pals_duration>(.+)<\/pals_duration>/) {
        $duration= 60 * $1;
	next;
      }
      if (/<pals_repeat>Y<\/pals_repeat>/) {
 	$bits|= 0x200;
 	next;
      }
      if (/<pals_captioning>Y<\/pals_captioning>/) {
 	$bits|= 1;
 	next;
      }
      if (/<pad_final>Y<\/pad_final>/) {
        $premiere= "Final";
	next;
      }
      if (/<pals_state>(.+)<\/pals_state>/) {
	$state=$1;
	next;
      }
      if (/<pals_year_produced>(.+)<\/pals_year_produced>/) {
        $year= $1;
	next;
      }
      if (/<pad_start_time>(.+)<\/pad_start_time>/) {
	Debug("Found start time $1\n");
        $time= TwenFourToSecs($1);

	# Naughty ABC lets programs start after 2359 :-)
	if ($time>=86400) {
	  $time-= 86400; $daynum++;
	  Debug("Bumped ABC daynum to $daynum\n");
 	}
	next;
      }
      if (/<pals_synopsis><..CDATA.(.+)]]><\/pals_synopsis>/) {
        $description= $1;

	# Trim off any parentheses at the end of the description
	$description=~ s/\(.*\)$//;
	$description=~ s/\s+$//;

	# Extract actors and director if they exist
	if ($description=~ /(.*)\s*CAST:\s*(.*)/) {
	  $description= $1; $actorline=$2;
	  if ($actorline=~ /(.*)\.*\s*DIR:\s*(.*)/) {
	    $actorline= $1; $director=$2;
	  }
	  $actor= $actorline;
	  $actor=~ s/,\s*/,/g;
	  $actor=~ s/\s*$//;
	}
	next;
      }
    }
  }
  close(IN);
  return(1);
}

1;
