#!/usr/bin/perl
# 
# AugmentData.pm: Augment data in intermediate files to speed up slice making.
#
# Copyright (c) Warren Toomey, wkt@tuhs.org, 2002/2003/2004, GPL license
#
#    $Revision: 3.21 $
#
# This file is distributed under the GNU Public License with the
# following additional limitations.
# 
# 1. This software must not be used in those parts of the world where
#    TiVo Inc. provides a subscription service.
# 2. This software must not be used for commercial purposes. You may 
#    not use this software to provide a "service" and charge for that
#    service.
#
# See Docs/COPYING for the GNU Public License details.
#
package TiVo::AugmentData;
use strict;
use TiVo::ConfigFiles;
use TiVo::MiscUtil;
use TiVo::ProgramDb;
use TiVo::Ratings;
use TiVo::Directors;
use TiVo::ProgramGenre;
use HTTP::Lite;
use CGI::Enurl;


BEGIN {
  use Exporter	 ();
  use vars	 qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

  $VERSION = do { my @r = (q$Revision: 3.21 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; # must be all one line, for MakeMaker

  @ISA	       = qw(Exporter);
  @EXPORT      = qw(Augment_Data);
  %EXPORT_TAGS = ( );
  @EXPORT_OK   = qw();
}

use vars	 @EXPORT_OK;
END { }	 # module clean-up code here (global destructor)


#####################################
###### Package-private Globals ######
#####################################


######################
###### Routines ######
######################

# Load data from intermediate files, find which programs
# don't have a genre, find their genre, rewrite the file.
#
sub Augment_Data($$$$)
{
  my ($chan,$snum,$enum,$write_files)= @_;
  my ($wanteday, $dayname, $infile, $outfile);
  my ($match, $newtitle, $genrelist, $is_episodic);
  my ($i, $isep, $title, $leftovertitle, $genrefromfile);
  my ($g, $trimg, $rating, $startitle, $result, $director);
  my (%AlreadyDone);

  $result=1;
  # We need the %Programs database to be loaded, but it
  # is possible that make_slice will run and call us
  # without loading the database. So we test and load it
  # if required.
  if (!defined(%Programs)) {
    Debug("Loading Programs database in Augment_Data!\n");
    Read_Programs;
  }

  # Do each day in turn
  for ($wanteday= $snum; $wanteday <= $enum; $wanteday++) {

    $dayname= Get_Dayname($wanteday);
    $infile= "$ConfigVar{Datadir}/$chan" . "_$dayname";
    $outfile= "$ConfigVar{Extradir}/$chan" . "_$dayname";

    # Don't re-augment if the outfile exists
    if (-f $outfile ) {
	Debug("Not re-augmenting $outfile, it already exists\n");
	next;
    }

    # If no data found, and we are only doing 1 day, return an error.
    # Otherwise, just loop so that we don't write out the file.
    unless (Read_IntermediateFile($infile)) {
      Debug("No data found to augment for $chan on $dayname\n");
      if ($snum==$enum) { return(0); } else { next; }
    }
    Debug("Read intermediate data from $infile\n");

    # Now work on each title in turn
    for ($i=0; $i <= $#ProgramData; $i++) {

      # Always lowercase the title, so even if we don't
      # find it, it will still look good.
      $title= $ProgramData[$i]{title};
      ($title)= LowercaseString($title);
      Debug("Title from the file is $title\n");
      $genrefromfile= $ProgramData[$i]{genre};
      $is_episodic=0; undef($genrelist);

      # See if we have some results for this title already.
      # Only non-episodic programs will be cached.
      if (defined($AlreadyDone{$title})) {
	Debug("We have already augmented $title\n");
	$ProgramData[$i]{title}= $AlreadyDone{$title}{title};
	$ProgramData[$i]{genre}= $AlreadyDone{$title}{genre};
	$ProgramData[$i]{starrating}= $AlreadyDone{$title}{starrating};
 	$ProgramData[$i]{director}= $AlreadyDone{$title}{director};
	next;
      }

      # Nothing in the cache, so let's try the trimtitles first
      Debug("About to Find_Trimtitle for $title\n");
      ($match, $newtitle, $trimg, $isep)= Find_Trimtitle($title);
      if ($match==1) {
	# Found something, store the genre found and the new title
	$genrelist= $trimg;
	$title= $newtitle;
	$is_episodic= $isep;
	Debug("Found a trimtitle of $title, genre $trimg\n");
      }

      # Now see if there is a per-channel genre and override
      # the trimtitles genre
      $genrelist= $StationGenre{$chan} if (defined($StationGenre{$chan}));

      # Now override this if we actually have genre from the data file
      $genrelist= $genrefromfile if (defined($genrefromfile));

      # Finally search the program database to get the definitive genre.
      Debug("About to Find_Program for $title\n");
      ($match, $newtitle, $leftovertitle, $g)= Find_Program($title);
      if ($match==1) {
	Debug("Found match, saving $newtitle vs. $title in Augment\n");
	$genrelist= $g;
	$title= $newtitle;
	$ProgramData[$i]{episode}= $leftovertitle if ($leftovertitle ne "");
	$is_episodic= 1;
      }

      # One final kludge, if the trimtitle or channel says it's a movie,
      # add this to the final genrelist.
      if (($trimg=~ /Movie/) || ($StationGenre{$chan}=~ /Movie/)) {
	if (! ($genrelist=~ /Movie/)) {
	  $genrelist = $genrelist . ",Movie,Movies";
	  Debug("Added movie to $title via the kludge\n");
	}
      }

      # If we have star ratings, search for a movie's star ratings
      # Do this for movies or for 90min+ programs that are not sport and
      # do not yet exist in the programs database.
      if (($genrelist=~ /Movie/) ||
	  (($ProgramData[$i]{duration}>=5400) && !($genrelist=~ /Sport/)
							    && ($match==0))) {
        Debug("About to find star rating for $title $genrelist\n");
        $rating= Find_Rating($title,$ProgramData[$i]{year});
        if (defined($rating)) {
          Debug("Found star rating $rating for $title\n");
          $ProgramData[$i]{starrating}= $rating;
	  if (! ($genrelist=~ /Movie/)) {
	    $genrelist = $genrelist . ",Movie,Movies";
	    Debug("Added movie to $title via the second kludge\n");
	  }
        }
        $director= Find_Director($title,$ProgramData[$i]{year});
        if (defined($director)) {
          Debug("Found director $director for $title\n");
          $ProgramData[$i]{director}= $director;
        }
      }

      # We can now save the stuff we found: title, genre, episodic
      Debug("Saving $i $title $is_episodic $genrelist\n");
      $ProgramData[$i]{title}= $title;
      $ProgramData[$i]{genre}= $genrelist;
      $ProgramData[$i]{episodic}= $is_episodic;

      # If we found a non-episodic program, cache it for later.
      # We don't cache episodic programs, as we want to get
      # different episode names.
      if (!defined($ProgramData[$i]{episodic}) || ($is_episodic==0)) {
	$AlreadyDone{$title}{title}= $ProgramData[$i]{title};
	$AlreadyDone{$title}{genre}= $ProgramData[$i]{genre};
        $AlreadyDone{$title}{starrating}= $ProgramData[$i]{starrating};
        $AlreadyDone{$title}{director}= $ProgramData[$i]{director};
      }
    }

    # All done, write the new file out.
    if ($write_files==1) {
      unless (Write_IntermediateFile($outfile)) {
        print(STDERR "Unable to write to file $infile\n");
	$result=0;
      }
    }
  }
  return($result);
}

1;
