#!/usr/bin/perl
# 
# AustarFetcher.pm: parse the data in web pages from www.austar.com.au
# and construct suitable data structures which can be converted into slice
# files for the TiVo.
#
# Due to legal reasons, I have had to cripple this fetcher so that it
# cannot grab data from the Foxtel website. All that is left is skeleton
# code which should be very similar to the code in all the other fetchers.
#
# Copyright (c) Warren Toomey, wkt@tuhs.org, 2002/2003/2004, GPL license
#
#    $Revision: 1.9 $
#
# This file is distributed under the GNU Public License with the
# following additional limitations.
# 
# 1. This software must not be used in those parts of the world where
#    TiVo Inc. provides a subscription service.
# 2. This software must not be used for commercial purposes. You may 
#    not use this software to provide a "service" and charge for that
#    service.
#
# See Docs/COPYING for the GNU Public License details.
#
package TiVo::AustarFetcher;
use strict;
use TiVo::ConfigFiles;
use TiVo::MiscUtil;
use HTTP::Lite;

BEGIN {
  use Exporter	 ();
  use vars	 qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

  $VERSION = do { my @r = (q$Revision: 1.9 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; # must be all one line, for MakeMaker

  @ISA	       = qw(Exporter);
  @EXPORT      = qw(Fetch_Austar_Channel);
  %EXPORT_TAGS = ( );
  @EXPORT_OK   = qw();
}

use vars	 @EXPORT_OK;
END { }	 # module clean-up code here (global destructor)


#####################################
###### Package-private Globals ######
#####################################

# Base URL used for Austar.
my ($AustarCookieUrl)=
	"";
my ($AustarDataUrl)=
	"";

# Words to look for in program description to find the actors
my (@Try_cast)= ("Starring", "Hosted by", "Presented by",
		 "Cast:", "Guest stars", "Stars");

# Which days have already been parsed
my (%AlreadyDoneDay);

# The array of program data that is
# built up.
my (@AustarData, $progindex);

######################
###### Routines ######
######################

# Given a daynumber relative to today (0, 1, 2 etc.), return the
# dayname as a Austar string.
sub Get_Austar_Dayname($)
{
  my ($wanteday)= $_[0];
  my ($dayname);
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst);

  # We want a yyyy_mm_dd format
  ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)= Get_Daylist($wanteday);
  $mon++; $year+= 1900;
  $dayname= sprintf("%04d_%d_%d", $year, $mon, $mday);
  return($dayname);
}

# Given a channel name, the relative start and
# end days, fetch the web data for this channel.
# Rehtml it and save it. Then parse the data
# into the intermediate format. Returns 1 on
# no errors, 0 if errors.
sub Fetch_Austar_Channel($$$)
{
  my ($chan,$snum,$enum)= @_;
  my ($result1,$result2)= (1,1);
  $result1=Fetch_and_Store_Data($chan,$snum,$enum);
  $result2=Parse_Data($chan,$snum,$enum);
  return($result1 && $result2);
}

sub Fetch_and_Store_Data($$$)
{
  my ($chan,$snum,$enum)= @_;
  my ($web_proxy, $webdelay, $wanteday);
  my ($html_file, $dayname, $daynum, $outfile);
  my ($result, $try_again, $try, $http, $req, $body);
  my (%args, $daystring, $cookie, $req);

  # We start 1 day early so that we get the midnight to 6am data
  $result=1; $snum--;

  # See if we need a web proxy, and # of times to try again
  $web_proxy= $ConfigVar{Webproxy};
  $try_again= $ConfigVar{RepeatWebConnect};

  # Set the delay in seconds between web fetches
  $webdelay=0;
  if (defined($ConfigVar{WebDelay})) {
    $webdelay= int($ConfigVar{WebDelay});
  }

  # Do each day in turn
  for ($wanteday= $snum; $wanteday <= $enum; $wanteday++) {

    $dayname= Get_Dayname($wanteday);
    $daynum= Get_Daynumber($wanteday);

    # Skip if we have already generated the output file
    $outfile=	"$ConfigVar{Datadir}/$chan" . "_$dayname";
    if (-f $outfile) {
      Debug("Intermediate file $outfile exists, not fetching\n");
      next;
    }

    # Get the output filename
    $html_file= "$ConfigVar{Webdir}/$chan" . "_$dayname";

    # If the HTML file already exists, assume we have already run
    # this before and stored it in the cache.
    if ( -f $html_file) {
      Debug("Using cached data for channel $chan for $dayname\n");
    } else {
      Debug("Fetching programs for channel $chan for $dayname\n");

      $http = new HTTP::Lite;
      if (defined($web_proxy)) { $http->proxy($web_proxy); }

      # Get the day string as required by Austar
      $daystring= Get_Austar_Dayname($wanteday);

      %args = ("date" => "$daystring",
		"searchtype" => "powersearch",
		"x" => 9, "y" => 17, "channel" => $StationUrl{$chan});

      # Try 1 or more times to get the file
      if (!defined($try_again)) { $try_again=1; }

      for ($try=$try_again;$try>0;$try--) {
      }
      $body= $http->body();
      Rehtml($body, $html_file, 0);
      WebLink($html_file, $chan, $dayname);
      sleep($webdelay);
    }
  }
  return($result);
}

sub Get_Austar_Cookie()
{
}

# WebLink: see if other channels use the same URL and
# make a symlink to the file instead of fetching it.
sub WebLink($$$)
{
  my ($file, $origchan, $dayname)= @_;
  my ($chan, $html_file);

  foreach $chan (keys(%StationUrl)) {
    next if ($chan eq $origchan);
    if ($StationUrl{$chan} eq $StationUrl{$origchan}) {
      $html_file= "$ConfigVar{Webdir}/$chan" . "_$dayname";
      link($file, $html_file);
      Debug("Linked $file to $html_file as same URL\n");
    }
  }
}


# Given the channel, start day and end day, read the webfiles and
# convert into intermediate format. We now put all of one day, and
# only that day, into an output file.
# New overall strategy: read in each file and parse it.
# Build up a list of records with no duration. Once the
# files are read in, go back and deduce the duration
# by looking at the differences between start times.
# Finally, write the files out.
#
sub Parse_Data($$$)
{
  my ($chan,$snum,$enum)= @_;
  my (%IgnoreDay);
  my ($wanteday, $outfile, $result);
  my ($dayname, $daynum, $i, $pdindex, $start_index);

  $result=1; $progindex=0; undef(@AustarData);

  # Find out what days we must actually do, and parse the web files.
  for ($wanteday= $snum; $wanteday <= $enum; $wanteday++) {
    $dayname= Get_Dayname($wanteday);
    $outfile=	"$ConfigVar{Datadir}/$chan" . "_$dayname";

    # Skip if we have already generated the output file
    if (-f $outfile) {
      Debug("Intermediate file $outfile exists, not overwriting\n");
      $IgnoreDay{$wanteday}=1;
      # We insert a fictitious record with a 6am start anyway
      $AustarData[$progindex]{title}="Fictitious 6am record";
      $AustarData[$progindex]{daynum}= $AustarData[$progindex-1]{daynum};
      $AustarData[$progindex]{time}=21600; $progindex++;
      next;
    }

    # Build the list of records for the previous day and this day
    if (!defined($AlreadyDoneDay{$wanteday-1})) {
      $result &= Parse_Data_For_Day($chan,$wanteday-1);
    }
    $result &= Parse_Data_For_Day($chan,$wanteday);
    $AlreadyDoneDay{$wanteday}=1;
  }

  # Insert a final record to allow the last duration calculation
  $AustarData[$progindex]{title}="Fictitious 6am record";
  $AustarData[$progindex]{daynum}= $AustarData[$progindex-1]{daynum};
  $AustarData[$progindex]{time}=21600; $progindex++;

  # Calculate the duration for each program
  for ($i=0; $i< $progindex -1; $i++) {
    $AustarData[$i]{duration}=
	$AustarData[$i+1]{time} - $AustarData[$i]{time};
    $AustarData[$i]{duration}+= 86400 if ($AustarData[$i]{duration}<0);
    # Debug("$i: $AustarData[$i]{title}\n");
    # Debug("$AustarData[$i]{daynum}\n");
    # Debug("$AustarData[$i]{time} $AustarData[$i]{duration}\n\n");
  }

  # Now write the results out for each day where there is no data yet
  $start_index=0;
  for ($wanteday= $snum; $wanteday <= $enum; $wanteday++) {
    if (defined($IgnoreDay{$wanteday})) { next; }

    # Print out the relevant records for this daynum
    $dayname= Get_Dayname($wanteday);
    $daynum=  Get_Daynumber($wanteday); $pdindex=0; undef(@ProgramData);
    $outfile=	"$ConfigVar{Datadir}/$chan" . "_$dayname";

    for ($i=$start_index; $i< $progindex -1; $i++) {

      next if ($AustarData[$i]{daynum}<$daynum);
      if ($AustarData[$i]{daynum}>$daynum) {
	$start_index=$i; last;
      }
      Debug("Adding $AustarData[$i]{title} to $daynum\n");
      $ProgramData[$pdindex]= $AustarData[$i]; $pdindex++;
    }
    if ($pdindex!=0) {
      Debug("Writing file $outfile\n");
      $result &= Write_IntermediateFile($outfile);
    } else {
      print(STDERR "No Austar data for $chan on $dayname\n");
      $result=0;
    }
  }
  return($result);
}


# Parse a web file for only a single day.
sub Parse_Data_For_Day($$)
{
  my ($chan,$wanteday)= @_;
  my ($dayname, $daynum);
  my ($html_file);
  my ($this_time, $last_start_time);
  my ($hr, $min, $apm, $start_time);
  my ($title);
  my ($i, @Actorlist, $description);
  my ($actor);
  my ($g);

  $dayname= Get_Dayname($wanteday);
  $daynum=  Get_Daynumber($wanteday);
  $last_start_time=-1;

  # Get the input filename
  $html_file= "$ConfigVar{Webdir}/$chan" . "_$dayname";

  unless (open(IN, $html_file)) {
    print(STDERR "Can't open downloaded web page $html_file: $!\n");
    return(0);
  }
  Debug("Parsing Austar data from $html_file\n");

  while (<IN>) {
  }
  close(IN);

  if ($last_start_time == -1) {
    Debug("No Austar data in $html_file\n");
    # Unlink the file as it contained no data
    unlink($html_file);
    return(0);
  }
  return(1);
}

# Procedure find_the_cast:
# Use heuristics to try & work out who is in the cast.
# Look for CAST, Stars, Starring, UPPERCASE WORDS (for Channel Ten).
#
sub Find_The_Cast($$) {
    my ($channel, $dref)= @_;
    my ($description)= $$dref;
    my (@Actor, $trycast, $left, $right, $actor);

    foreach $trycast (@Try_cast) {
      #
      # If we find a match using any of the trycast words,
      # followed by a : or a space, then ....
      #
      if ($description =~ /(.*)$trycast(:\s*|\s+)/i) {
        $left=$1;
        #
        # Remove the stuff leading up to and including the trycast.
        # Then remove anything starting with a dot, a '(' and
        # words starting with lowercase letters.
        #
        $description=~ s/.*$trycast\s*:*\s*//i;
        $description=~ s/((\b[a-z].*|\..*|\(.*))//;

        if (defined($1)) { $right = $1; } else { $right = ""; }
        $$dref = $left . $right;	# Send new description back.
        # 
        # Now split on either commas or the patterns ' and '
        # and ' & '
        #
        push(@Actor,
      	LowercaseString(split(/,\s*|\s+and\s+|\s*&\s+/,$description)));
        return(@Actor);
      }
    }

    # No, luck try the uppercase method then :-)
    #
    #while ($description =~ /([A-Z][A-Z]+ [A-Z][A-Z]+)/g) {
    #  ($actor)= LowercaseString($1);
    #  Debug("Working on $actor\n");
    #  push(@Actor,$actor);
    #}
    return(@Actor);
}

1;
