#!/usr/bin/perl
# 
# GenreSearch.pm: Routines to search for unknown genres in intermediate data.
#
# Copyright (c) Warren Toomey, wkt@tuhs.org, 2002/2003/2004, GPL license
#
#    $Revision: 3.23 $
#
# This file is distributed under the GNU Public License with the
# following additional limitations.
# 
# 1. This software must not be used in those parts of the world where
#    TiVo Inc. provides a subscription service.
# 2. This software must not be used for commercial purposes. You may 
#    not use this software to provide a "service" and charge for that
#    service.
#
# See Docs/COPYING for the GNU Public License details.
#
package TiVo::GenreSearch;
use strict;
use TiVo::ConfigFiles;
use TiVo::MiscUtil;
use TiVo::ProgramGenre;
use TiVo::ProgramDb;
use TiVo::Ratings;
use HTTP::Lite;
use CGI::Enurl;


BEGIN {
  use Exporter	 ();
  use vars	 qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

  $VERSION = do { my @r = (q$Revision: 3.23 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; # must be all one line, for MakeMaker

  @ISA	       = qw(Exporter);
  @EXPORT      = qw(List_Unknown_Genres Search_Unknown_Genres);
  %EXPORT_TAGS = ( );
  @EXPORT_OK   = qw();
}

use vars	 @EXPORT_OK;
END { }	 # module clean-up code here (global destructor)


#####################################
###### Package-private Globals ######
#####################################

my (%UnknownProgram, %Program, %AlreadyCreated);

# Hashes to convert Yahoo! genres to TiVo genres
my (%Genre) = (
  "Action" => "Action,ActionAdventureGroup",
  "Adventure" => "Action,ActionAdventureGroup",
  "Animation" => "Animated,ChildrensGroup",
  "Arts and Culture" => "Art,ArtsMusicLiving",
  "Biography" => "Biography,DocumentaryGroup",
  "Business and Finance" => "Business,NewsBusinessWeather",
  "Cartoon" => "Animated,ChildrensGroup",
  "Children" => "Animated,ChildrensGroup",
  "Classic" => "Classic",
  "Comedy" => "Comedy,ComedyGroup",
  "Crime" => "Crime,DramaGroup",
  "Current Affairs" => "PublicAffairs,NewsBusinessWeather",
  "Documentary" => "Documentary,DocumentaryGroup",
  "Drama" => "Drama,DramaGroup",
  "Education" => "Educational,EducationalGroup",
  "Entertainment" => "Unknown",
  "Family" => "Family,ChildrensGroup",
  "Game Show" => "GameShow,Daytime",
  "Historical" => "Historical,DocumentaryGroup",
  "Horror" => "Horror,Thriller",
  "Infotainment" => "Shopping",
  "Lifestyle" => "Outdoors,RealLife",
  "Music" => "Music,ArtsMusicLiving",
  "Musical" => "Musical,ArtsMusicLiving",
  "Mystery" => "Mystery,MysteryThriller",
  "N/A" => "Unknown",
  "Nature" => "Nature,DocumentaryGroup",
  "News" => "News,NewsBusinessWeather",
  "Real Life" => "RealLife",
  "Religion" => "Religion",
  "Romance" => "Romance,DramaGroup",
  "Romantic Comedy" => "RomanticComedy,ComedyGroup",
  "Sci-Fi" => "ScienceFiction,SciFiFantasy",
  "Science and Technology" => "Science,ScienceNatureTechnology",
  "Shopping" => "Shopping,Daytime",
  "Soap Opera" => "SoapOpera,DramaGroup",
  "Sitcom" => "Situation,Daytime",
  "Sport" => "Sports,SportsGroup",
  "Talk Show" => "Talkshow,Daytime",
  "Thriller" => "Thriller",
  "Travel" => "Travel",
  "Variety" => "Variety",
  "War" => "War,DocumentaryGroup",
  "Western" => "Western"
);


######################
###### Routines ######
######################

# Load data from intermediate files, find which programs
# don't have a genre, and build a list of them
#
sub List_Unknown_Genres($$$)
{
  my ($chan,$snum,$enum)= @_;
  my ($wanteday, $dayname, $infile);
  my ($match, $newtitle, $genrelist, $is_episodic);
  my ($i, $title, $leftovertitle, $shortday, $rating);
  my (%TryList);

  # Do each day in turn
  for ($wanteday= $snum; $wanteday <= $enum; $wanteday++) {

    $dayname= Get_Dayname($wanteday);
    $shortday= substr($dayname,0,3);
    $infile= "$ConfigVar{Datadir}/$chan" . "_$dayname";
    unless (Read_IntermediateFile($infile)) {
      # print(STDERR "Can't open data file $infile: $!\n");
      next;
    }
    Debug("Read intermediate data from $infile\n");

    for ($i=0; $i <= $#ProgramData; $i++) {

      # Skip this program if it has a genre with the word Movie in it
      if (defined($ProgramData[$i]{genre}) &&
	  ($ProgramData[$i]{genre}=~ /Movie/)) {
        Debug("Skip $ProgramData[$i]{title}, it's a movie\n");
	next;
      }

      # Do the same if no genre but it is on a Movie channel
      if (defined($StationGenre{$chan}) && ($StationGenre{$chan}=~ /Movie/)) {
        Debug("Skip $ProgramData[$i]{title}, it's on a movie channel\n");
	next;
      }

      # Get the details of this program
      $title= $ProgramData[$i]{title};
      ($title)= LowercaseString($title);
      $TryList{$title}{channel}= $chan;
      $TryList{$title}{day}= $shortday;
      $TryList{$title}{time}=$ProgramData[$i]{time};
      $TryList{$title}{duration}=$ProgramData[$i]{duration};
      $TryList{$title}{year}=$ProgramData[$i]{year};
      # Record the genre, if any.
      if (defined($ProgramData[$i]{genre}) && !defined($TryList{$title}{genre})){
        Debug("Saving genre $ProgramData[$i]{genre} for $title\n");
        $TryList{$title}{genre}=$ProgramData[$i]{genre};
      }
    }
  }

  # We have built a list of all titles. Now see which ones
  # are not in the programs database,
  foreach $title (sort keys %TryList) {
    Debug("Working on >$title<\n");
    # Start with a match on the trimtitles
    #
    ($match, $newtitle, $genrelist, $is_episodic)= Find_Trimtitle($title);
    if ($match==1) { next; }

    # No trimtitles match, we have to try the programs database now
    #
    ($match, $newtitle, $leftovertitle, $genrelist)= Find_Program($title);
    if ($match==1) {
      Debug("Program match found on $title\n");
      next;
    }

    # Kludge: try to find a program which is 90mins+ and is defined
    # in the ratings database. If we find one, it's probably a
    # movie. Skip it.
    if ($TryList{$title}{duration}>=5400) {
      $rating= Find_Rating($title,$TryList{title}{year});
      if (defined($rating)) {
        Debug("Found movie using movie kludge on $title\n");
        next;
      }
    }

    # We didn't find a match in the programs file and it's not a movie.
    # If we do have a genre defined, then assume that it is valid, and
    # store it into the programs database.
    if (defined($TryList{$title}{genre}) && !defined($AlreadyCreated{$title})) {
      Debug("Adding $title to db as genre $TryList{$title}{genre}\n");
      $Programs{$title}{title}=$title;
      $Programs{$title}{channel}=$TryList{$title}{channel};
      $Programs{$title}{time}=$TryList{$title}{time};
      $Programs{$title}{day}=$TryList{$title}{day};
      $Programs{$title}{duration}=$TryList{$title}{duration};
      $Programs{$title}{genre}=$TryList{$title}{genre};
      Create_Program($title);
      $AlreadyCreated{$title}=1;
    } else {
      #
      # Nope, no genre, add it to the list of programs to search for
      # If we have a station genre, use it here.
      #
      print("Program $title unknown\n");
      $UnknownProgram{$title}{channel}=$TryList{$title}{channel};
      $UnknownProgram{$title}{time}=$TryList{$title}{time};
      $UnknownProgram{$title}{day}=$TryList{$title}{day};
      $UnknownProgram{$title}{duration}=$TryList{$title}{duration};
      $UnknownProgram{$title}{genre}= $StationGenre{$chan}
				if (defined($StationGenre{$chan}));
    }
  }
  undef(%TryList);
}


# Search the list of unknown genre programs
sub Search_Unknown_Genres {
  my ($body, $enctitle, $event, $found, $genre);
  my ($http, $shorttitle, $partmatch, $req, $title);
  my ($trytitle, %AlreadyDone);
  my ($web_proxy, $webdelay);
  my($tmpfile, $tmpfile2);
  my $Tempdir= $ConfigVar{Tempdir};

  $Tempdir= "/tmp" if (!defined($Tempdir));
  $tmpfile= "$Tempdir/ya.$$";
  $tmpfile2= "$Tempdir/ya2.$$";

  $web_proxy= $ConfigVar{Webproxy};

  # Set the delay in seconds between web fetches
  $webdelay=0;
  if (defined($ConfigVar{WebDelay})) {
    $webdelay= int($ConfigVar{WebDelay});
  }

  foreach $title (sort keys(%UnknownProgram)) {
    $shorttitle= $title;
    $partmatch=0;
    # Get rid of stuff after last colon
    if ($shorttitle=~ /: [^:]*$/) {
      $shorttitle=~ s/: [^:]*$//;
      # $partmatch=1;
    }
    # Ditto for " - "
    if ($shorttitle=~ / - .*$/) {
      $shorttitle=~ s/ - .*$//;
      # $partmatch=1;
    }

    if ($AlreadyDone{$title}) { next; }
    print("Searching for $title");
    $found=0;
  
    # Convert title into URL encoded form, and lowercase it for later
    $enctitle= enurl($shorttitle);
    ($title)= LowercaseString($title);
  
    # Now send the request to search for this title
    $http = new HTTP::Lite;
    $http->add_req_header("Referer", "http://au.tv.yahoo.com/");

    if (defined($web_proxy)) { $http->proxy($web_proxy); }
    $req = $http->request("http://au.tv.yahoo.com/searchresults.html?" .
  		"kw=$enctitle&x=32&y=8") || die "Unable to search $title: $!";
    if ($req ne "200") {
      print(STDERR "Search for $title failed ($req): ".$http->status_message());
      next;
    }
    $body = $http->body();
  
    # Save and convert the incoming HTML to one tag per line
    Rehtml($body,$tmpfile,0);
  
    # Build up some of the final record now
    $Programs{$title}{partial}=$partmatch;
    $Programs{$title}{title}=$title;
    $Programs{$title}{channel}=$UnknownProgram{$title}{channel};
    $Programs{$title}{time}=$UnknownProgram{$title}{time};
    $Programs{$title}{day}=$UnknownProgram{$title}{day};
    $Programs{$title}{duration}=$UnknownProgram{$title}{duration};

    # Now walk the file, getting event URLs and the following titles
    open(HIN, $tmpfile) || die("Cannot open temp file $tmpfile\n");
    while(<HIN>) {
      if (/event.html\?ev=(\d+)\"/) {
        $event= $1;
        $trytitle= <HIN>;
        chop($trytitle);
        ($trytitle)= LowercaseString($trytitle);
  
        # Do a match on the title and the trytitle, both lowercased
        if ($title eq $trytitle) {
  	# Now fetch that web page, given the event number
    	$http = new HTTP::Lite;
        if (defined($web_proxy)) { $http->proxy($web_proxy); }
    	$http->add_req_header("Referer", "http://au.tv.yahoo.com/");
    	$req = $http->request("http://au.tv.yahoo.com/event.html?ev=$event")
  					|| die "Unable to get document: $!";
    	if ($req ne "200") {
      	  print(STDERR "Request failed ($req): ".$http->status_message());
	  close(HIN); next;
    	}
    	$body = $http->body();
  
    	# Save and convert the incoming HTML to one tag per line
    	Rehtml($body,$tmpfile2,0);
  
  	# Find the yahoo genre and print it and the original title to stdout
  	open(GIN, $tmpfile2) || die("Cannot open $tmpfile2");
  	while(<GIN>) {
  	  chop;
  	  if (/Genre:\s*(.*)/) {
  	    $genre=$1; $found=1;
  	    if (defined($Genre{$genre})) {
  	      print(", found genre $Genre{$genre}\n");
	      $Programs{$title}{genre}=$Genre{$genre};
	      # Don't add movies!
	      if (! ($Programs{$title}{genre}=~ /^Movie/) ) {
      		Debug("Adding $title to db as genre $Genre{$genre}, #1\n");
	        Create_Program($title) if (!defined($AlreadyCreated{$title}));
      		$AlreadyCreated{$title}=1;
	      }
  	    } else {
  	      print(", unknown yahoo genre $genre\n");
	      if (defined($UnknownProgram{$title}{genre})) {
		$Programs{$title}{genre}= $UnknownProgram{$title}{genre};
	      } else {
                $Programs{$title}{genre}= "Unknown";
	      }
	      if (! ($Programs{$title}{genre}=~ /^Movie/) ) {
      		Debug("Adding $title to db as genre $Genre{$genre}, #2\n");
	        Create_Program($title) if (!defined($AlreadyCreated{$title}));
      		$AlreadyCreated{$title}=1;
	      }
  	    }
  	    last;
  	  }
  	}
  	close(GIN); last;
        }
      }
    }
    close(HIN);
    if ($found==0) {
      print(", not found");
      if (defined($UnknownProgram{$title}{genre})) {
	$Programs{$title}{genre}= $UnknownProgram{$title}{genre};
        print(", using $UnknownProgram{$title}{genre}");
      } else {
        $Programs{$title}{genre}= "Unknown";
      }
      print("\n");
      if (! ($Programs{$title}{genre}=~ /^Movie/) ) {
        Debug("Adding $title to db as genre $Genre{$genre}, #3\n");
	Create_Program($title) if (!defined($AlreadyCreated{$title}));
     	$AlreadyCreated{$title}=1;
      }
    }
    
    # Now sleep for $webdelay seconds so as not to annoy yahoo
    $AlreadyDone{$title}=1;
    sleep($webdelay);
  }
  
  # All done
  close(HIN); unlink($tmpfile);
  unlink($tmpfile2);
}

1;
