#!/usr/bin/perl
# 
# PlusNFetcher: A pseudo-fetchers which reads data from existing
# intermediate files and adds N hours to it. This is for those
# channels which are simply N hours delayed from the original channels.
#
# Copyright (c) Warren Toomey, wkt@tuhs.org, 2004, GPL license
#
#    $Revision: 1.3 $
#
# This file is distributed under the GNU Public License with the
# following additional limitations.
# 
# 1. This software must not be used in those parts of the world where
#    TiVo Inc. provides a subscription service.
# 2. This software must not be used for commercial purposes. You may 
#    not use this software to provide a "service" and charge for that
#    service.
#
# See Docs/COPYING for the GNU Public License details.
#
package TiVo::PlusNFetcher;
use strict;
use TiVo::ConfigFiles;
use TiVo::MiscUtil;

BEGIN {
  use Exporter	 ();
  use vars	 qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

  $VERSION = do { my @r = (q$Revision: 1.3 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; # must be all one line, for MakeMaker

  @ISA	       = qw(Exporter);
  @EXPORT      = qw(Fetch_PlusN_Channel);
  %EXPORT_TAGS = ( );
  @EXPORT_OK   = qw();
}

use vars	 @EXPORT_OK;
END { }	 # module clean-up code here (global destructor)


#####################################
###### Package-private Globals ######
#####################################

######################
###### Routines ######
######################

# Given a channel name, the relative start and
# end days, read the intermediate data for the
# channel, alter the times for the programs
# and save it to new intermediate data files.
# Returns 1 on no errors, 0 if errors.
sub Fetch_PlusN_Channel($$$)
{
  my ($chan,$snum,$enum)= @_;
  my ($url,$oldchan,$offset,$daynum);
  my ($wanteday,$dayname,$infile,$outfile);
  my (%OrigData,$i,$cnt,$result);

  # Split our "URL" into the old station-name and the offset. There
  # must be a "+" sign in the URL somewhere.
  $url= $StationUrl{$chan};
  if (! ($url=~ /\+/)) {
    print(STDERR "Error: PlusNfetcher need + sign in URL for $chan\n");
    return(0);
  }

  ($oldchan, $offset)= split(/\+/, $url);
  $offset= 3600 * $offset;

  # Fetch all the intermediate data from the original channels
  # from $snum-1 up to $enum
  for ($wanteday= $snum-1; $wanteday <= $enum; $wanteday++) {

    $dayname= Get_Dayname($wanteday);
    $daynum=  Get_Daynumber($wanteday);
    $infile= "$ConfigVar{Datadir}/$oldchan" . "_$dayname";

    # Ignore read errors for now. Copy results array.
    if (Read_IntermediateFile($infile)) {
      @{ $OrigData{$wanteday} } = @ProgramData;

      # Alter the times for all the records
      # Add the offset on to the start time and alter the day if needed
      for ($i=0; $i <= $#{ $OrigData{$wanteday} }; $i++) {
        $OrigData{$wanteday}[$i]{time}+= $offset;
        if ($OrigData{$wanteday}[$i]{time}>86400) {
	  $OrigData{$wanteday}[$i]{time}-= 86400;
	  $OrigData{$wanteday}[$i]{daynum}++;
        }
      }
    }
  }

  # Now find and add those records which have moved to our day
  $result=1;
  for ($wanteday= $snum; $wanteday <= $enum; $wanteday++) {
    undef(@ProgramData); $cnt=0;
    $dayname= Get_Dayname($wanteday);
    $daynum=  Get_Daynumber($wanteday);

    # Do the previous day first
    for ($i=0; $i <= $#{ $OrigData{$wanteday-1} }; $i++) {
      # Use it if it's on our day
      if ($OrigData{$wanteday-1}[$i]{daynum}==$daynum) {
	$ProgramData[$cnt]= $OrigData{$wanteday-1}[$i]; $cnt++;
	Debug("Added $OrigData{$wanteday-1}[$i]{title} to list for $daynum\n");
      }
    }

    # Now do our day, should be the bulk of the data
    for ($i=0; $i <= $#{ $OrigData{$wanteday} }; $i++) {
      # Use it if it's on our day
      if ($OrigData{$wanteday}[$i]{daynum}==$daynum) {
	$ProgramData[$cnt]= $OrigData{$wanteday}[$i]; $cnt++;
	Debug("Added $OrigData{$wanteday}[$i]{title} to list for $daynum\n");
      }
    }

    # Write the new file out
    $outfile=   "$ConfigVar{Datadir}/$chan" . "_$dayname";

    # Skip if we have already generated the output file
    if (-f $outfile) {
      Debug("Intermediate file $outfile exists, not overwriting\n");
      next;
    }

    # Skip if there was no original data for today
    if ($#{ $OrigData{$wanteday} }<0) {
      Debug("No PlusN data for $outfile\n");
      next;
    }

    Debug("Writing file $outfile\n");
    $result &= Write_IntermediateFile($outfile);
  }
  return(1);
}

1;
