#!/usr/bin/perl
# 
# ProgramGenre.pm: Routines to use the programs database, and to determine
# the genre of a given program.
#
# Copyright (c) Warren Toomey, wkt@tuhs.org, 2002/2003/2004, GPL license
#
#    $Revision: 3.6 $
#
# This file is distributed under the GNU Public License with the
# following additional limitations.
# 
# 1. This software must not be used in those parts of the world where
#    TiVo Inc. provides a subscription service.
# 2. This software must not be used for commercial purposes. You may 
#    not use this software to provide a "service" and charge for that
#    service.
#
# See Docs/COPYING for the GNU Public License details.
#
package TiVo::ProgramGenre;
use strict;
use TiVo::ConfigFiles;
use TiVo::MiscUtil;
use TiVo::ProgramDb;

BEGIN {
  use Exporter	 ();
  use vars	 qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

  $VERSION = do { my @r = (q$Revision: 3.6 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; # must be all one line, for MakeMaker

  @ISA	       = qw(Exporter);
  @EXPORT      = qw(Find_Trimtitle Find_Program);
  %EXPORT_TAGS = ( );
  @EXPORT_OK   = qw();
}

use vars	 @EXPORT_OK;
END { }	 # module clean-up code here (global destructor)


######################
###### Routines ######
######################

# Apply the TrimTitle list to the given title
# and if there is a match, return the new title,
# the genre and if the program is episodic.
sub Find_Trimtitle($)
{
  my ($title)= @_;
  my ($trimtitle, $proglen, $lctitle);
  my ($match, $newtitle, $genrelist, $is_episodic);
  $match=0;

  # Convert the title to lowercase for the comparison
  # and coalesce whitespace
  $lctitle= lc($title);
  $lctitle=~ s/\s+/ /g;

  # Trim any part of the title from the beginning, and
  # perhaps get an idea of the genre from this bit.
  # Also remove leading whitespace, : and -
  foreach $trimtitle (keys(%TrimTitle)) {
    if ( $lctitle =~ /^\Q$trimtitle\E/i ) {
      $proglen= length($trimtitle);
      $newtitle= substr($title, $proglen);
      $newtitle=~ s/^[-\s:]+//;
      $genrelist=$TrimTitle{$trimtitle}{genre};
      $is_episodic= $TrimTitle{$trimtitle}{episodic};
      $match=1; last;
    }
  }
  return($match, $newtitle, $genrelist, $is_episodic);
}

# Apply the %Programs and %Partialprograms lists
# to the title to find out its genre. If there is
# a match, return the correct title, anything leftover
# from the title and the genre.
sub Find_Program($)
{
  my ($title)= @_;
  my ($match, $newtitle, $leftovertitle, $genrelist, $lctitle, $i);

  # Convert the title to lowercase for the comparison
  # and replace/coalesce punctuation & whitespace
  $lctitle= lc($title);
  $lctitle=~ y/-:;?.'/      /;
  $lctitle=~ s/\s+/ /g;

  # Try to find an exact match
  if (defined($Programs{$lctitle})) {
    Debug("Found exact match for $title\n");
    $newtitle = $Programs{$lctitle}{title};
    $leftovertitle= "";
    $genrelist= $Programs{$lctitle}{genre};
    return(1,$newtitle,$leftovertitle,$genrelist);
  }

  # An exact match failed, so we try a partial match now
  #
  $match=0;
  foreach $i (@SortedPrograms) {
    if ( $lctitle =~ /^\Q$i\E/ ) {
      Debug("Found partial match $i vs $title\n");
      $newtitle = $PartialPrograms{$i}{title};
      # Trim off the title to get the leftover title. At
      # the same time remove leading whitespace, : and -
      $leftovertitle= substr($title, length($i));
      $leftovertitle=~ s/^[-\s:]+//;
      Debug("Set leftover $leftovertitle for $title\n");
      $genrelist= $PartialPrograms{$i}{genre};
      $match=1; last;
    }
  }
  if ($match==0) { Debug("No programs match for $title\n"); }
  return($match,$newtitle,$leftovertitle,$genrelist);
}

1;
